--[[

$Revision: 144 $

(C) Copyright 2009 Bethink (bethink at naef dot com)
See LICENSE.txt for license terms.

]]


----------------------------------------------------------------------
-- Initialization

BossNotesAbilities = LibStub("AceAddon-3.0"):NewAddon("BossNotesAbilities",
	"AceEvent-3.0",
	"AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BossNotes")


----------------------------------------------------------------------
-- Constants

-- Database version
BOSS_NOTES_ABILITIES_DATABASE_VERSION = 2

-- Provider name
BOSS_NOTES_ABILITIES_PROVIDER = "ABILITIES"

-- Flags
BOSS_NOTES_ABILITIES_DIFFICULTY = 255
BOSS_NOTES_ABILITIES_NORMAL = 1
BOSS_NOTES_ABILITIES_HEROIC = 2
BOSS_NOTES_ABILITIES_USAGE = 65280
BOSS_NOTES_ABILITIES_CAST = 256
BOSS_NOTES_ABILITIES_AURA = 1536
BOSS_NOTES_ABILITIES_BUFF = 512
BOSS_NOTES_ABILITIES_DEBUFF = 1024

-- Ace configuration options
local ACE_OPTS = {
	name = L["ABILITIES"],
	handler = BossNotesAbilities,
	type = "group",
	args = {
		tooltipInfo = {
			name = L["TOOLTIP_INFO"],
			desc = L["TOOLTIP_INFO_DESC"],
			order = 1,
			type = "toggle",
			get = "IsTooltipInfo",
			set = "SetTooltipInfo"
		},
		learn = {
			name = L["LEARN"],
			desc = L["LEARN_DESC"],
			order = 2,
			type = "toggle",
			get = "IsLearning",
			set = "ToggleLearning"
		},
		clear = {
			name = L["CLEAR"],
			desc = L["CLEAR_DESC"],
			order = 3,
			type = "execute",
			func = "ClearLearned"
		}
	}
}
BossNotes:MergeOptions("abilities", ACE_OPTS)

-- DB defaults
local DB_DEFAULTS = {
	global = {
		tooltipInfo = true,
		npcs = { },
		instanceEncounters = { }
	}
}

-- Tracked events
local EVENTS = {
	SPELL_DAMAGE = BOSS_NOTES_ABILITIES_CAST,
	SPELL_MISSED =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_ENERGIZE =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_DRAIN =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_LEECH =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_EXTRA_ATTACKS =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_CAST_START =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_CAST_SUCCESS =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_CAST_FAILED =  BOSS_NOTES_ABILITIES_CAST,
	SPELL_AURA_APPLIED =  BOSS_NOTES_ABILITIES_AURA,
	SPELL_AURA_REMOVED = BOSS_NOTES_ABILITIES_AURA
	--SPELL_AURA_APPLIED_DOSE = BOSS_NOTES_ABILITIES_AURA,
	--SPELL_AURA_REMOVED_DOSE = BOSS_NOTES_ABILITIES_AURA,
	--SPELL_AURA_REFRESH = BOSS_NOTES_ABILITIES_AURA
}

-- Filtered NPCs
local FILTERED_NPCS = {
	[510] = true,   -- Water Elemental
	[2630] = true,  -- Earthbind Totem
	[15438] = true, -- Greater Fire Elemental
	[19668] = true, -- Shadowfiend
	[24207] = true, -- Army of the Dead Ghoul
	[26125] = true, -- Risen Ghoul
	[27829] = true, -- Ebon Gargoyle
	[27893] = true, -- Rune Weapon
	[28306] = true, -- Anti-Magic Zone
	[29264] = true, -- Spirit Wolf
	[29998] = true, -- Desecrated Ground V
	[31216] = true, -- Mirror Image
	[31167] = true, -- Magma Totem VII
	[33753] = true  -- Desecrated Ground IV
}

	
----------------------------------------------------------------------
-- Core handlers

-- One-time initialization
function BossNotesAbilities:OnInitialize ()
	-- Database
	self.db = LibStub("AceDB-3.0"):New("BossNotesAbilitiesDB")
	self.db:RegisterDefaults(DB_DEFAULTS)
	self:MigrateDatabase(BOSS_NOTES_ABILITIES_DATABASE_VERSION)
	
	-- Prune learned that are now in the encounter database
	for key, instanceEncounter in pairs(self.db.global.instanceEncounters) do
		local empty = true
		for npcId in pairs(instanceEncounter.npcIds) do
			if BossNotes:GetInstanceAndEncounter(npcId) then
				instanceEncounter.npcIds[npcId] = nil
			else
				empty = false
			end
		end
		if empty then 
			self.db.global.instanceEncounters[key] = nil
		end
	end
end

-- Handles the (re-)enabling of this add-on
function BossNotesAbilities:OnEnable ()
	-- Register provider
	BossNotes:RegisterProvider(BOSS_NOTES_ABILITIES_PROVIDER, {
		OnSources = function (instance, encounter) return self:OnSources(instance, encounter) end
	})

	-- Register events
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	
	-- Hooks
	self:HookScript(GameTooltip, "OnTooltipSetUnit")
end

-- Handles the disabling of this add-on
function BossNotesAbilities:OnDisable ()
	-- Unregister provider
	BossNotes:UnregisterProvider(BOSS_NOTES_ABILITIES_PROVIDER)
end


----------------------------------------------------------------------
-- BossNotes provider

-- Returns the abilities note
function BossNotesAbilities:OnSources (instance, encounter)
	-- Get and sort NPC IDs
	local npcIds = (encounter and encounter.npcIds) or (instance and instance.npcIds)
	local key = BossNotes:GetContextKey(instance, encounter)
	local instanceEncounter = self.db.global.instanceEncounters[key]
	if instanceEncounter then
		-- Merge
		local mergedNpcIds = { }
		if npcIds then
			for _, npcId in ipairs(npcIds) do
				if not instanceEncounter.npcIds[npcId] then
					table.insert(mergedNpcIds, npcId)
				end
			end
		end
		for npcId in pairs(instanceEncounter.npcIds) do
			table.insert(mergedNpcIds, npcId)
		end
		npcIds = mergedNpcIds
	end
	if not npcIds then
		return
	end
	local encounterId = encounter and encounter.encounterId
	table.sort(npcIds, function (a, b)
		if encounterId then
			if a == encounterId then
				return true
			end
			if b == encounterId then
				return false
			end
		end
		local npcA = self.db.global.npcs[a]
		local npcB = self.db.global.npcs[b]
		if npcA and npcB then
			return npcA.name < npcB.name
		end
		return npcA and not npcB
	end)
	
	-- Build HTML
	local html = { }
	local difficulty = self:GetDifficulty(instance and instance.raid)
	table.insert(html, "<html><body>")
	for _, npcId in ipairs(npcIds) do
		local npc = self.db.global.npcs[npcId]
		if npc then
			-- Get and sort spells
			local spells = { }
			for _, spell in pairs(npc.spells) do
				if bit.band(spell.flags, difficulty) ~= 0 then
					table.insert(spells, spell)
				end
			end
			
			-- Add NPC to HTML if there is at least one spell
			if #spells > 0 then
				table.sort(spells, function (a, b)
					return a.name < b.name
				end)
				table.insert(html, "<h1>" .. npc.name .. "</h1>")
				for _, spell in ipairs(spells) do
					local line = string.format(L["SPELL"], spell.spellId, spell.name)
					local usage = { }
					if bit.band(spell.flags, BOSS_NOTES_ABILITIES_CAST) ~= 0 then
						table.insert(usage, L["CAST"])
					end
					if bit.band(spell.flags, BOSS_NOTES_ABILITIES_BUFF) ~= 0 then
						table.insert(usage, L["BUFF"])
					end
					if bit.band(spell.flags, BOSS_NOTES_ABILITIES_DEBUFF) ~= 0 then
						table.insert(usage, L["DEBUFF"])
					end
					if #usage > 0 then
						line = line .. " (" .. table.concat(usage, ", ") .. ")"
					end
					table.insert(html, "<p>" .. line .. "</p>")
				end
				table.insert(html, "<br />")
			end
		end
	end
	if #html == 1 then
		table.insert(html, "<p>" .. L["ABILITIES_NOT_RECORDED"] .. "</p>")
	end
	table.insert(html, "</body></html>")
	
	return {
		title = L["ABILITIES"],
		content = table.concat(html, "\n")
	}
end


----------------------------------------------------------------------
-- Events

-- Handles a combat log event
function BossNotesAbilities:COMBAT_LOG_EVENT_UNFILTERED (message, timestamp, event,
		sourceGuid, sourceName, sourceFlags, destGuid, destName, destFlags, ...)
	-- Check event
	local usage = EVENTS[event]
	if not usage then
		return
	end
	
	-- Check unit affiliation
	if bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_MASK) ~= COMBATLOG_OBJECT_AFFILIATION_OUTSIDER then
		return
	end

	-- Check unit NPC ID
	local npcId = BossNotes:GetNpcId(sourceGuid)
	if not npcId or FILTERED_NPCS[npcId] then
		return
	end
	local instance, encounter = BossNotes:GetInstanceAndEncounter(npcId)
	if not instance then
		if self.learningZone and self.learningZone == GetRealZoneText() then
			instance, encounter = BossNotes:GetSelection()
			local key = BossNotes:GetContextKey(instance, encounter)
			local instanceEncounter = self.db.global.instanceEncounters[key]
			if not instanceEncounter then
				instanceEncounter = {
					contextText = BossNotes:GetContextText(instance, encounter),
					npcIds = { }
				}
				self.db.global.instanceEncounters[key] = instanceEncounter
			end
			if not instanceEncounter.npcIds[npcId] then
				BossNotes:Print(string.format(L["LEARNING"],
						sourceName, npcId, instanceEncounter.contextText))
				instanceEncounter.npcIds[npcId] = sourceName
				BossNotes:InvalidateContextKey(key)
			end
		else
			return
		end
	end
	
	-- Track NPC
	local npc = self.db.global.npcs[npcId]
	if not npc then
		npc = {
			name = sourceName,
			spells = { }
		}
		self.db.global.npcs[npcId] = npc
	end
	
	-- Track spell
	local spellId = select(1, ...)
	local spellName = select(2, ...)
	local spell = npc.spells[spellId]
	if not spell then
		spell = {
			spellId = spellId,
			name = spellName,
			flags = 0
		}
		npc.spells[spellId] = spell
	end
	
	-- Track difficulty and usage
	if usage == BOSS_NOTES_ABILITIES_AURA then
		usage = 0
		local auraType = select(4, ...)
		if auraType == "BUFF" then
			usage = BOSS_NOTES_ABILITIES_BUFF
		elseif auraType == "DEBUFF" then
			usage = BOSS_NOTES_ABILITIES_DEBUFF
		end
	end
	local flags = bit.bor(self:GetDifficulty(instance.raid), usage)
	if bit.band(spell.flags, flags) ~= flags then
		spell.flags = bit.bor(spell.flags, flags)
		BossNotes:InvalidateContextKey(BossNotes:GetContextKey(instance, encounter))
	end
end

-- Handles chat messages
function BossNotesAbilities:CHAT_MSG_SYSTEM (message, msg)
	if self:MatchFormat(msg, ERR_DUNGEON_DIFFICULTY_CHANGED_S) or
		self:MatchFormat(msg, ERR_RAID_DIFFICULTY_CHANGED_S) then
		BossNotes:InvalidateContextKey("all")
	end
end


----------------------------------------------------------------------
-- Hooks

-- Adds the abilities to the tooltip
function BossNotesAbilities:OnTooltipSetUnit (gameTooltip)
	if self.db.global.tooltipInfo then 
		local npcId = BossNotes:GetNpcId(UnitGUID("mouseover"))
		if npcId then
			local npc = self.db.global.npcs[npcId]
			if npc then
				-- Get current dungeon difficulty bit
				local _, instanceType = GetInstanceInfo()
				local difficulty = self:GetDifficulty(instanceType == "raid")

				-- Acquire and sort spell names
				local spellNameSet = { }
				for _, spell in pairs(npc.spells) do
					if bit.band(spell.flags, difficulty) ~= 0 then
						spellNameSet[spell.name] = true
					end
				end
				spellNames = { }
				for spellName in pairs(spellNameSet) do
					table.insert(spellNames, spellName)
				end
				table.sort(spellNames, function (a, b)
					return a < b
				end)
				
				-- Add to tooltip
				local abilities = table.concat(spellNames, ", ")
				local pos = 1
				abilities = string.gsub(abilities, "%s+()(%S+)()", function (start, word, stop)
					if stop - pos > 40 then
						pos = start
						return "\n" .. word
					end
				end)
				gameTooltip:AddLine(abilities, 0.91, 0.44, 0.0)
			end
		end
	end
end


----------------------------------------------------------------------
-- Options

-- Returns whether the tooltip information is enabled
function BossNotesAbilities:IsTooltipInfo (info)
	return self.db.global.tooltipInfo
end

-- Toggles the tooltip information
function BossNotesAbilities:SetTooltipInfo (info, tooltipInfo)
	self.db.global.tooltipInfo = tooltipInfo
end

-- Returns whether the module is learning about NPCs
function BossNotesAbilities:IsLearning (info)
	return self.learningZone
end

-- Toggles learning about NPCs
function BossNotesAbilities:ToggleLearning (info)
	if not self.learningZone then
		self.learningZone = GetRealZoneText()
		BossNotes:Print(string.format(L["START_LEARNING"], self.learningZone))
	else
		self.learningZone = nil
		BossNotes:Print(L["STOP_LEARNING"])
	end
end

-- Clears learned instance encounters
function BossNotesAbilities:ClearLearned ()
	-- Clear
	local instanceEncounters = self.db.global.instanceEncounters
	self.db.global.instanceEncounters = { }
	BossNotes:Print(L["CLEAR_LEARNED"])
	
	-- Invalidate
	for key in pairs(instanceEncounters) do
		BossNotes:InvalidateContextKey(key)
	end
end


----------------------------------------------------------------------
-- Utilities

-- Migrates the database
function BossNotesAbilities:MigrateDatabase (targetVersion)
	local version = self.db.global.databaseVersion or targetVersion
	
	-- Migrate version 1 -> 2
	if version == 1 and targetVersion >= 2 then
		for _, npc in pairs(self.db.global.npcs) do
			for _, spell in pairs(npc.spells) do
				if not spell.flags then
					spell.flags = spell.difficulty
				end
				spell.difficulty = nil
			end
		end
		version = 2
		self.db.global.databaseVersion = 2
	end
	
	-- Set to target version
	self.db.global.databaseVersion = version
end

-- Returns the dungeon difficulty bit
function BossNotesAbilities:GetDifficulty (raid)
	local rawDifficulty = IsInInstance() and GetInstanceDifficulty()
			or raid and GetRaidDifficulty() or GetDungeonDifficulty()
	return bit.lshift(1, rawDifficulty - 1)
end

-- Matches a format
function BossNotesAbilities:MatchFormat (msg, fmt)
	-- Create pattern
	local pattern = string.gsub(fmt, "%.", "%.")
	pattern = string.gsub(pattern, "%%s", "(.+)")
	pattern = string.gsub(pattern, "%%d", "(%d+)")
	pattern = "^" .. pattern .. "$"
	
	-- Match
	local found, _, capture = string.find(msg, pattern)
	return found, capture
end
