--[[

$Revision: 126 $

(C) Copyright 2009 Bethink (bethink at naef dot com)
See LICENSE.txt for license terms.

]]


----------------------------------------------------------------------
-- Initialization

BossNotesPersonalNotes = LibStub("AceAddon-3.0"):NewAddon("BossNotesPersonalNotes",
	"AceEvent-3.0",
	"AceHook-3.0",
	"AceSerializer-3.0",
	"AceComm-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("BossNotes")

	
----------------------------------------------------------------------
-- Constants

-- Provider name
BOSS_NOTES_PERSONAL_NOTES_PROVIDER = "PERSONAL_NOTES"

-- Database version
BOSS_NOTES_PERSONAL_NOTES_DATABASE_VERSION = 1

-- Comm prefix
BOSS_NOTES_PREFIX = "BossNotes"

-- Cache TTL
local CACHE_TTL = 300

-- Limits
local TITLE_LENGTH = 60
local CONTENT_LENGTH = 2000
local PRIVATE_CONTENT_LENGTH = 10000

-- Empty notes
local EMPTY_NOTES = {
	[""] = { }
}

-- Checks for validating incoming messages
local KEY_PATTERNS = {
	"^encounter:%d+$",
	"^instance:[%a%d]+$",
	"^general$"
}
local CHECKS = {
	request = {
		key = {
			type = "string",
			func = function (key)
				for _, pattern in ipairs(KEY_PATTERNS) do
					if string.match(key, pattern) then return true end
				end
				return false
			end
		}
	},
	response = {
		key = {
			type = "string",
			func = function (key)
				for _, pattern in ipairs(KEY_PATTERNS) do
					if string.match(key, pattern) then return true end
				end
				return false
			end
		},
		class = {
			type = "string",
			func = function (class) return RAID_CLASS_COLORS[class] end
		},
		title = {
			type = "string",
			func = function (title) return string.len(title) <= TITLE_LENGTH end
		},
		content = {
			type = "string",
			func = function (content) return string.len(content) <= CONTENT_LENGTH end
		},
		time = {
			type = "number",
			func = function (time) return time >= 0 end
		},
		clientVersion = {
			type = "string",
			func = function (clientVersion) return string.match(clientVersion, "^%d+%.%d+%.%d+$") end
		}
	}
}

-- Ace configuration options
local ACE_OPTS = {
	name = L["SYNDICATION"],
	desc = L["SYNDICATION_DESC"],
	handler = BossNotesPersonalNotes,
	type = "group",
	args = {
		guild = {
			name = L["GUILD"],
			order = 1,
			arg = "syndicationGuild",
			type = "toggle",
			get = "GetSyndication",
			set = "SetSyndication"
		},
		party = {
			name = L["PARTY"],
			order = 2,
			arg = "syndicationParty",
			type = "toggle",
			get = "GetSyndication",
			set = "SetSyndication"
		},
		raid = {
			name = L["RAID"],
			order = 3,
			arg = "syndicationRaid",
			type = "toggle",
			get = "GetSyndication",
			set = "SetSyndication"
		}
	}
}
BossNotes:MergeOptions("syndication", ACE_OPTS)

-- DB defaults
local DB_DEFAULTS = {
	global = {
		syndicationGuild = true,
		syndicationParty = true,
		syndicationRaid = true
	},
	char = {
		notes = {
		}
	}
}

-- Static poup dialog for adding a note
StaticPopupDialogs["BOSS_NOTES_PERSONAL_NOTES_ADD_NOTE"] = {
	text = L["ENTER_NOTE_NAME_PROMPT"],
	button1 = ACCEPT,
	button2 = CANCEL,
	hasEditBox = 1,
	maxLetters = TITLE_LENGTH,
	OnAccept = function (self)
		local text = strtrim(self.editBox:GetText())
		BossNotesPersonalNotes:SetNote(BossNotesPersonalNotes.contextSource.noteKey, text, nil)
		BossNotesPersonalNotes:SetNotePrivate(BossNotesPersonalNotes.contextSource.noteKey, text,
				BossNotesPersonalNotes.contextPrivate)
	end,
	EditBoxOnEnterPressed = function (self)
		local text = strtrim(self:GetParent().editBox:GetText())
		local notes = BossNotesPersonalNotes:GetNotes(BossNotesPersonalNotes.contextSource.noteKey)
		if not notes[text] then
			BossNotesPersonalNotes:SetNote(BossNotesPersonalNotes.contextSource.noteKey, text, nil)
			BossNotesPersonalNotes:SetNotePrivate(BossNotesPersonalNotes.contextSource.noteKey, text,
					BossNotesPersonalNotes.contextPrivate)
			self:GetParent():Hide()
		end
	end,
	EditBoxOnEscapePressed = function (self)
		self:GetParent():Hide()
	end,
	EditBoxOnTextChanged = function (self)
		local text = strtrim(self:GetParent().editBox:GetText())
		local notes = BossNotesPersonalNotes:GetNotes(BossNotesPersonalNotes.contextSource.noteKey)
		if not notes[text] then
			self:GetParent().button1:Enable();
		else
			self:GetParent().button1:Disable();
		end
	end,
	OnShow = function (self)
		self.button1:Disable();
		self.button2:Enable();
		self.editBox:SetFocus()
	end,
	OnHide = function (self)
		if ChatFrameEditBox:IsShown() then
			ChatFrameEditBox:SetFocus()
		end
		self.editBox:SetText("")
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}

-- Static popup dialog for confirming the removal of a note
StaticPopupDialogs["BOSS_NOTES_PERSONAL_NOTES_REMOVE_NOTE"] = {
	text = L["REMOVE_NOTE_PROMPT"],
	button1 = YES,
	button2 = NO,
	OnAccept = function (self)
		BossNotesPersonalNotes:RemoveNote(BossNotesPersonalNotes.contextSource.noteKey,
				BossNotesPersonalNotes.contextSource.noteTitle)
	end,
	OnCancel = function (self) end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}


----------------------------------------------------------------------
-- Core handlers

-- One-time initialization
function BossNotesPersonalNotes:OnInitialize()
	-- Player name and class
	self.playerName = UnitName("player")
	_, self.playerClass = UnitClass("player")
	
	-- Database
	self.db = LibStub("AceDB-3.0"):New("BossNotesPersonalNotesDB")
	self.db:RegisterDefaults(DB_DEFAULTS)
	self.db.global.databaseVersion = BOSS_NOTES_PERSONAL_NOTES_DATABASE_VERSION
	if BossNotes.db.char.notes then
		-- Migration
		for key, note in pairs(BossNotes.db.char.notes) do
			note.content = note.note
			note.note = nil
			self.db.char.notes[key] = {
				[""] = note
			}
		end
		BossNotes.db.char.notes = nil
	end

	-- Cache
	self.caches = { }
end

-- Handles (re-)enabling of the addon
function BossNotesPersonalNotes:OnEnable ()
	self:RegisterComm(BOSS_NOTES_PREFIX)
	self:SecureHook("ChatEdit_InsertLink", "InsertLink")
	BossNotes:RegisterProvider(BOSS_NOTES_PERSONAL_NOTES_PROVIDER, {
		OnSources = function (instance, encounter) return self:OnSources(instance, encounter) end,
		OnRefresh = function (instance, encounter) self:OnRefresh(instance, encounter) end,
		OnClick = function (instance, encounter, source, button) self:OnClick(instance, encounter, source, button) end
	})
end

-- Handles disabling of the addon
function BossNotesPersonalNotes:OnDisable ()
	BossNotes:UnregisterProvider(BOSS_NOTES_PERSONAL_NOTES_PROVIDER)
end


----------------------------------------------------------------------
-- Event handlers

-- Handles the Boss Notes sources request
function BossNotesPersonalNotes:OnSources (instance, encounter)
	local sources = { }
	
	-- Add personal notes
	local key = BossNotes:GetContextKey(instance, encounter)
	local notes = self:GetNotes(key)
	for title, note in pairs(notes) do
		local noteTitle = title
		if string.len(title) > 0 then
			title = self.playerName .. " - " .. title
		else
			title = self.playerName
		end
		table.insert(sources, {
			order = BOSS_NOTES_ORDER_PERSONAL,
			color = RAID_CLASS_COLORS[self.playerClass],
			title = title,
			tag = note.private and L["PRIVATE"],
			content = note.content,
			placeholder = L["CLICK_TO_EDIT"],
			time = note.time,
			clientVersion = note.clientVersion,
			private = note.private,
			noteKey = key,
			noteTitle = noteTitle
		})
	end
	
	-- Add syndicated notes from cache
	local cache = self:GetCache(key)
	for name, player in pairs(cache.players) do
		for title, note in pairs(player.notes) do
			if string.len(title) > 0 then
				title = name .. " - " .. title
			else
				title = name
			end
			table.insert(sources, {
				order = BOSS_NOTES_ORDER_COMMUNITY,
				color = RAID_CLASS_COLORS[player.class],
				title = title,
				content = note.content,
				time = note.time,
				clientVersion = note.clientVersion
			})
		end
	end
	
	-- Refresh asnychronously if TTL has expired
	self:RequestSources(instance, encounter, false)
	
	return sources
end

-- Handles the Boss Notes refresh request
function BossNotesPersonalNotes:OnRefresh (instance, encounter)
	self:RequestSources(instance, encounter, true)
end

-- Handles the Boss Notes click notification
function BossNotesPersonalNotes:OnClick (instance, encounter, source, button)
	if button == "LeftButton" then
		if source.noteKey and source.noteTitle then
			self:ShowEditor(instance, encounter, source,
					source.private and PRIVATE_CONTENT_LENGTH or CONTENT_LENGTH)
		end
	elseif button == "RightButton" then
		-- Setup
		self.contextSource = source
		if self.contextSource.noteKey and self.contextSource.noteTitle
				and string.len(self.contextSource.noteTitle) > 0 then
			self.removeNoteButton:Enable()
		else
			self.removeNoteButton:Disable()
		end

		-- Show
		local menu = BossNotesPersonalNotesContextMenu
		menu:Hide()
		local x, y = GetCursorPosition()
		menu:ClearAllPoints()
		menu:SetPoint("TOPLEFT", "UIParent", "BOTTOMLEFT", x, y)
		menu:Show()
		UIMenu_StartCounting(menu)
	end
end

-- Handles the loading of the context menu
function BossNotesPersonalNotes:OnLoadContextMenu (frame)
	frame:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
	frame:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
	UIMenu_Initialize(frame)
	UIMenu_AddButton(frame, L["ADD_NOTE"], nil, function ()
		self.contextPrivate = nil
		StaticPopup_Show("BOSS_NOTES_PERSONAL_NOTES_ADD_NOTE")
	end)
	UIMenu_AddButton(frame, L["ADD_PRIVATE_NOTE"], nil, function ()
		self.contextPrivate = true
		StaticPopup_Show("BOSS_NOTES_PERSONAL_NOTES_ADD_NOTE")
	end)
	UIMenu_AddButton(frame, L["REMOVE_NOTE"], nil, function ()
		StaticPopup_Show("BOSS_NOTES_PERSONAL_NOTES_REMOVE_NOTE",
				self.contextSource.title)
	end)
	UIMenu_FinishInitializing(frame)
	self.removeNoteButton = _G[frame:GetName() .. "Button3"]
end


----------------------------------------------------------------------
-- Hooks

-- Inserts a link into the Boss Notes editor
function BossNotesPersonalNotes:InsertLink (link)
	local editBox = BossNotesPersonalNotesEditorScrollFrameEditBox
	if editBox:IsVisible() and editBox:HasFocus() then
		editBox:Insert(link)
	end
end


----------------------------------------------------------------------
-- Notes management

-- Returns the personal notes for a key
function BossNotesPersonalNotes:GetNotes (key)
	if self.db.char.notes[key] then
		return self.db.char.notes[key]
	else
		return EMPTY_NOTES
	end
end

-- Sets a personal note
function BossNotesPersonalNotes:SetNote (key, title, content)
	-- Normalize content to nil if empty
	if content then
		content = strtrim(content)
		if string.len(content) == 0 then
			content = nil
		end
	end
	
	-- Set or clear, invalidating sources as needed
	local modified = false
	local notes = self.db.char.notes[key]
	if not notes then
		notes = { }
		self.db.char.notes[key] = notes
	end
	local note = notes[title]
	if note then
		if content ~= note.content then
			modified = true
		end
	else
		note = { }
		notes[title] = note
		modified = true
	end
	if modified then
		note.content = content
		if content then
			note.time = time()
			note.clientVersion = GetBuildInfo()
		else
			note.time = nil
			note.clientVersion = nil
		end
		BossNotes:InvalidateContextKey(key, true)
	end
end

-- Sets whether a note is private
function BossNotesPersonalNotes:SetNotePrivate (key, title, private)
	if private then
		private = true
	else
		private = nil
	end
	if self.db.char.notes[key][title].private ~= private then
		self.db.char.notes[key][title].private = private
		BossNotes:InvalidateContextKey(key, true)
	end
end
	
-- Removes a personal note
function BossNotesPersonalNotes:RemoveNote (key, title)
	-- Cannot remove default note
	if string.len(title) == 0 then
		return
	end
	
	-- Defined?
	local notes = self.db.char.notes[key]
	if not notes or not notes[title] then
		return
	end
	
	-- Remove
	notes[title] = nil
	BossNotes:InvalidateContextKey(key, true)
	
	-- Close open editor
	if self.editSource and self.editSource.noteKey == key
			and self.editSource.noteTitle == title then
		if BossNotesPersonalNotesEditor:IsShown() then
			BossNotesPersonalNotesEditor:Hide()
		end
	end
end


----------------------------------------------------------------------
-- Communications

-- Requests the sources on the network
function BossNotesPersonalNotes:RequestSources (instance, encounter, force)
	-- Request from network if TTL has expired
	local key = BossNotes:GetContextKey(instance, encounter)
	local cache = self:GetCache(key)
	local time = time()
	if time - cache.lastRequest > CACHE_TTL or force then
		-- Create message
		local request = {
			type = "request",
			key = key
		}
		local message = self:Serialize(request)
		
		-- Send
		if self.db.global.syndicationGuild and IsInGuild() then
			self:SendCommMessage(BOSS_NOTES_PREFIX, message, "GUILD")
		end
		if self.db.global.syndicationParty and UnitInParty("player") then
			self:SendCommMessage(BOSS_NOTES_PREFIX, message, "PARTY")
		end
		if self.db.global.syndicationRaid and UnitInRaid("player") then
			self:SendCommMessage(BOSS_NOTES_PREFIX, message, "RAID")
		end
	end
end

-- Handles comm messages
function BossNotesPersonalNotes:OnCommReceived (prefix, message, distribution, sender)
	-- Check distribution
	if not (distribution == "GUILD" and self.db.global.syndicationGuild or
		distribution == "PARTY" and self.db.global.syndicationParty or
		distribution == "RAID" and self.db.global.syndicationRaid) then
		return
	end
	
	-- Deserialize and validate
	local success, object = self:Deserialize(message)
	if not success or not self:CheckObject(object) then
		return
	end
	object.distribution = distribution
	object.sender = sender
	
	-- Process
	if object.type == "request" then
		self:OnCommReceivedRequest(object)
	elseif object.type == "response" then
		self:OnCommReceivedResponse(object)
	end
end

-- Handles a request
function BossNotesPersonalNotes:OnCommReceivedRequest (object)
	-- Note time of request
	local cache = self:GetCache(object.key)
	cache.lastRequest = time()
	
	-- Send response
	local notes = self:GetNotes(object.key)
	for title, note in pairs(notes) do
		if note.content and not note.private then
			-- Create message
			local response = {
				type = "response",
				key = object.key,
				class = self.playerClass,
				title = title,
				content = note.content,
				time = note.time,
				clientVersion = note.clientVersion
			}
			local message = self:Serialize(response)
			self:SendCommMessage(BOSS_NOTES_PREFIX, message, object.distribution)
		end
	end
end

-- Handles a response
function BossNotesPersonalNotes:OnCommReceivedResponse (object)
	-- Ignore our own response
	if object.sender == self.playerName then
		return
	end
	
	-- Cache response
	local modified = false
	local player = self:GetPlayer(self:GetCache(object.key), object.sender)
	if not player.class then
		player.class = object.class
		modified = true
	end
	local note = player.notes[object.title]
	if note then
		modified = modified or object.content ~= note.content
				or object.time ~= note.time
				or object.clientVersion ~= note.clientVersion
	else
		note = { }
		player.notes[object.title] = note
		modified = true
	end
	if modified then
		-- Update
		note.content = object.content
		note.time = object.time
		note.clientVersion = object.clientVersion
		
		-- Invalidate
		BossNotes:InvalidateContextKey(object.key)
	end
end

-- Returns a per-key cache
function BossNotesPersonalNotes:GetCache (key)
	local cache = self.caches[key]
	if not cache then
		cache = {
			lastRequest = 0,
			players = { }
		}
		self.caches[key] = cache
	end
	return cache
end

-- Returns a player from a cache
function BossNotesPersonalNotes:GetPlayer (cache, name)
	local player = cache.players[name]
	if not player then
		player = {
			notes = { }
		}
		cache.players[name] = player
	end
	return player
end
			
-- Checks an incoming object
function BossNotesPersonalNotes:CheckObject (object)
	-- Is it a table?
	if type(object) ~= "table" then
		return false
	end
	
	-- Known type?
	local checks = CHECKS[object.type]
	if not checks then
		return false
	end

	-- Run checks
	for name, check in pairs(checks) do
		if type(object[name]) ~= check.type then
			return false
		end
		if not check.func(object[name]) then
			return false
		end
	end
	return true
end


----------------------------------------------------------------------
-- Options

-- Gets a global option
function BossNotesPersonalNotes:GetSyndication (info)
	return self.db.global[info.arg]
end

-- Sets a syndication option
function BossNotesPersonalNotes:SetSyndication (info, value)
	self.db.global[info.arg] = value
end
